@@ -92,7 +92,7 @@ function chunked_transmit(data: string, chunk_size: number = 8096): void {
 
 
 /**
- * Start the keepalive pinger to prevent nginx timeout
+ * Start the keepalive pinger to prevent nginx timeout using native WebSocket ping/pong
  */
 export function startKeepalivePinger(): void {
   // Clear any existing pinger first
@@ -101,38 +101,32 @@ export function startKeepalivePinger(): void {
     keepaliveInterval = null;
   }
 
-  console.log('🏓 Starting keepalive pinger');
-
   // Send ping every 50 seconds (10 seconds before nginx 60s timeout)
-  keepaliveInterval = setInterval(async () => {
+  keepaliveInterval = setInterval(() => {
     // Check connection state before pinging
     if (!ws || ws.readyState !== WebSocket.OPEN || !isConnected) {
-      console.log('🏓 Connection lost, attempting reconnection...');
       stopKeepalivePinger();
 
-      try {
-        await reconnectWebSocket();
-        console.log('🏓 Reconnection successful, pinger restarted');
-      } catch (error) {
+      reconnectWebSocket().then(() => {
+        // Reconnection successful, pinger restarted
+      }).catch((error) => {
         console.error('🏓 Reconnection failed:', error);
-      }
+      });
       return;
     }
 
     try {
-      // Use callPython to send a simple ping message
-      await callPython({
-        method: 'ping',
-        message_type: 'request',
-        call_id: generateUUID(),
-        timestamp: Date.now()
-      } as any, 5000, false, false); // 5s timeout, no response wait, no stop button
-
-      console.log('🏓 Sent keepalive ping');
+      // Use native WebSocket ping frame - no application-level message needed
+      (ws as any).ping?.(); // Some WebSocket implementations have ping() method
       reconnectAttempts = 0; // Reset reconnect attempts on successful ping
     } catch (error) {
-      console.log('🏓 Keepalive ping failed:', error);
-      // Don't immediately reconnect on ping failure - wait for next cycle
+      // Fallback: send a minimal data frame to keep connection alive
+      try {
+        ws.send(''); // Send empty string as keepalive
+        reconnectAttempts = 0;
+      } catch (sendError) {
+        // Connection is likely dead, let the next cycle handle reconnection
+      }
     }
   }, PING_INTERVAL);
 }
@@ -144,7 +138,6 @@ export function stopKeepalivePinger(): void {
   if (keepaliveInterval) {
     clearInterval(keepaliveInterval);
     keepaliveInterval = null;
-    console.log('🏓 Keepalive pinger stopped');
   }
 }
 
@@ -189,11 +182,8 @@ export async function reconnectWebSocket(): Promise<void> {
  * @returns Resolved URL with proper JupyterHub user path if needed
  */
 function resolveWebSocketURL(url: string): string {
-  console.log(`##>>## Original URL: ${url}`);
-
   // If it's already an absolute URL (starts with ws:// or wss://), return as-is
   if (url.startsWith('ws://') || url.startsWith('wss://')) {
-    console.log(`##>>## Absolute URL, using as-is: ${url}`);
     return url;
   }
 
@@ -211,10 +201,6 @@ function resolveWebSocketURL(url: string): string {
     const host = window.location.host;
     const fullURL = `${protocol}//${host}${resolvedPath}`;
 
-    console.log(`##>>## JupyterHub detected, resolved URL: ${fullURL}`);
-    console.log(`##>>## User path: ${userPath}`);
-    console.log(`##>>## Resolved path: ${resolvedPath}`);
-
     return fullURL;
   } else {
     // Standalone JupyterLab - construct normal URL
@@ -222,8 +208,6 @@ function resolveWebSocketURL(url: string): string {
     const host = window.location.host;
     const fullURL = `${protocol}//${host}${url}`;
 
-    console.log(`##>>## Standalone JupyterLab, resolved URL: ${fullURL}`);
-
     return fullURL;
   }
 }
@@ -266,18 +250,10 @@ export function initPythonBridge(url: string = "ws://127.0.0.1:8777/ws"): Promis
     // Check if we need to add bonnieUrl as query parameter for backend proxy
     const currentSettings = (window as any).escobarCurrentSettings;
     if (currentSettings?.bonnieUrl && url.startsWith('/')) {
-      console.log(`>>> Adding bonnieUrl as proxy target: ${currentSettings.bonnieUrl}`);
-
       // Add bonnieUrl as query parameter for backend to use as proxy target
       const separator = resolvedURL.includes('?') ? '&' : '?';
       resolvedURL += `${separator}bonnie_url=${encodeURIComponent(currentSettings.bonnieUrl)}`;
-
-      console.log(`>>> Final URL with bonnieUrl parameter: ${resolvedURL}`);
-    } else {
-      console.log(">>> No bonnieUrl parameter needed");
     }
-
-    console.log(">>> Connecting to " + resolvedURL);
     ws = new WebSocket(resolvedURL);
 
     ws.addEventListener('open', (event) => {
@@ -324,9 +300,22 @@ export function initPythonBridge(url: string = "ws://127.0.0.1:8777/ws"): Promis
       }
     });
 
+    // Add pong event listener for native WebSocket ping/pong
+    (ws as any).addEventListener?.('pong', () => {
+      console.log('🏓 Received native WebSocket pong response');
+      reconnectAttempts = 0; // Reset reconnect attempts on successful ping/pong
+    });
+
     ws.addEventListener('message', async (event: any) => {
       //const messageStr = typeof data === 'string' ? data : data.toString('utf-8');
       const messageStr = event.data;
+
+      // Skip processing empty messages (keepalive frames)
+      if (!messageStr || messageStr === '') {
+        console.log('🏓 Received keepalive response (empty message)');
+        return;
+      }
+
       try {
         var jsonData: any = messageStr;
         if (typeof messageStr === 'string') {
@@ -471,8 +460,8 @@ export async function callPython(message: import('../types/protocol').ProtocolMe
     originalMessage: message
   });
 
-  // Only dispatch start event for operations that should show the stop button
-  if (showStopButton) {
+  // Only dispatch start event for operations that wait for responses AND should show the stop button
+  if (waitForResponse && showStopButton) {
     console.log(`[callPython] ${timestamp} Dispatching START event for stop button`);
     window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.START));
   }
@@ -622,9 +611,11 @@ export async function callPython(message: import('../types/protocol').ProtocolMe
         pendingRequests.delete(call_id);
         console.log(`[callPython] ${timestamp} Removed from pending (was present: ${wasInPending}), remaining pending: ${pendingRequests.size}`);
 
-        // Dispatch end event on timeout
-        console.log(`[callPython] ${timestamp} Dispatching END event due to timeout for call_id: ${call_id}`);
-        window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.END));
+        // Only dispatch end event on timeout if we were waiting for a response AND showing stop button
+        if (waitForResponse && showStopButton) {
+          console.log(`[callPython] ${timestamp} Dispatching END event due to timeout for call_id: ${call_id}`);
+          window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.END));
+        }
 
         reject(new Error(`Request timed out after ${timeoutMs}ms`));
       }, timeoutMs);
@@ -641,19 +632,17 @@ export async function callPython(message: import('../types/protocol').ProtocolMe
               value
             });
 
-            // Only dispatch END event if finish_reason is NOT "tool_calls"
+            // Only dispatch END event if showStopButton is true AND finish_reason is NOT "tool_calls"
             // Hide stop button when: finish_reason is undefined/null OR finish_reason !== "tool_calls"
             const shouldHideStopButton = !value.finish_reason || value.finish_reason !== "tool_calls";
-            console.log(`[callPython] ${timestamp} Stop button decision for call_id: ${call_id}`, {
-              finishReason: value.finish_reason,
-              shouldHideStopButton
-            });
 
-            if (shouldHideStopButton) {
-              console.log(`[callPython] ${timestamp} Dispatching END event (hiding stop button) for call_id: ${call_id}`);
+            if (shouldHideStopButton && showStopButton) {
+              console.log(`[callPython] ${timestamp} Dispatching END event due to normal completion for call_id: ${call_id}`);
               window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.END));
+            } else if (shouldHideStopButton && !showStopButton) {
+              console.log(`[callPython] ${timestamp} Normal completion but no stop button was shown for call_id: ${call_id}`);
             } else {
-              console.log(`[callPython] ${timestamp} NOT dispatching END event (keeping stop button) for call_id: ${call_id}`);
+              console.log(`[callPython] ${timestamp} Tool calls continue - keeping stop button visible for call_id: ${call_id}`);
             }
 
             // TODO: Handle multiple responses in sequence - may need to track conversation state
@@ -671,9 +660,11 @@ export async function callPython(message: import('../types/protocol').ProtocolMe
               stack: reason instanceof Error ? reason.stack : 'N/A'
             });
 
-            // Always hide stop button on errors
-            console.log(`[callPython] ${timestamp} Dispatching END event due to rejection for call_id: ${call_id}`);
-            window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.END));
+            // Only hide stop button on errors if it was shown
+            if (showStopButton) {
+              console.log(`[callPython] ${timestamp} Dispatching END event due to error for call_id: ${call_id}`);
+              window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.END));
+            }
             reject(reason);
           },
           timeout
@@ -683,8 +674,7 @@ export async function callPython(message: import('../types/protocol').ProtocolMe
       } else {
         console.log(`[callPython] ${timestamp} Not waiting for response (waitForResponse=false), cleaning up immediately for call_id: ${call_id}`);
         pendingRequests.delete(call_id);
-        console.log(`[callPython] ${timestamp} Dispatching END event (no response wait) for call_id: ${call_id}`);
-        window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.END));
+        // No stop button events for operations that don't wait for responses
       }
 
       // Send the message
@@ -705,8 +695,10 @@ export async function callPython(message: import('../types/protocol').ProtocolMe
         // Clean up on send failure
         clearTimeout(timeout);
         pendingRequests.delete(call_id);
-        console.log(`[callPython] ${timestamp} Dispatching END event due to send failure for call_id: ${call_id}`);
-        window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.END));
+        if (showStopButton) {
+          console.log(`[callPython] ${timestamp} Dispatching END event due to send failure for call_id: ${call_id}`);
+          window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.END));
+        }
         reject(sendError);
       }
     } else {
@@ -719,9 +711,11 @@ export async function callPython(message: import('../types/protocol').ProtocolMe
         errorCondition: 'WebSocket not connected after attempting to reconnect'
       });
 
-      // Dispatch end event on immediate rejection
-      console.log(`[callPython] ${timestamp} Dispatching END event due to connection failure for call_id: ${call_id}`);
-      window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.END));
+      // Only dispatch end event on immediate rejection if stop button was shown
+      if (showStopButton) {
+        console.log(`[callPython] ${timestamp} Dispatching END event due to connection failure for call_id: ${call_id}`);
+        window.dispatchEvent(new CustomEvent(PYTHON_CALL_EVENTS.END));
+      }
       reject(new Error("Cannot send message. WebSocket is not connected after attempting to reconnect."));
     }
   });
